/*
 * @(#) $Id: mapupdate.c,v 1.1 2004/06/14 13:51:08 yeti Exp $
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2004 Yeti (David Necas) <yeti@physics.muni.cz>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "mapupdate.h"

enum { bpp = 3 };

void
update_map_square(gint width, gint height, gdouble *som,
                  gint m, gdouble w, gdouble *vec,
                  gint size, gdouble *kernel,
                  gint depth, TreeLevel **bbtree,
                  gboolean use_somtree)
{
  gint i, j, ii, jj;
  gint imin, imax, jmin, jmax;
  gint ch;
  gdouble ww;
  gdouble *p;

  i = m/width;
  j = m % width;
  imin = MAX(i - size, 0);
  imax = MIN(i + size, height-1);
  jmin = MAX(j - size, 0);
  jmax = MIN(j + size, width-1);
  for (ii = imin; ii <= imax; ii++) {
    for (jj = jmin; jj <= jmax; jj++) {
      ww = w*kernel[(ii + size - i)*(2*size + 1) + (jj + size - j)];
      p = som + (ii*width + jj)*bpp;
      for (ch = 0; ch < 3; ch++) {
        p[ch] += ww*vec[ch];
        p[ch] /= (1 + ww);
      }
    }
  }

  if (!use_somtree)
    return;

  update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                        jmin, imin, jmax+1, imax+1);
}

void
update_map_h_cylinder(gint width, gint height, gdouble *som,
                      gint m, gdouble w, gdouble *vec,
                      gint size, gdouble *kernel,
                      gint depth, TreeLevel **bbtree,
                      gboolean use_somtree)
{
  gint i, j, ii, jj;
  gint imin, imax, jmin, jmax, jmv;
  gint ch;
  gdouble ww;
  gdouble *p;

  i = m/width;
  j = m % width;
  imin = MAX(i - size, 0);
  imax = MIN(i + size, height-1);
  jmin = j - size;
  jmax = j + size;
  jmv = (1 + size/width)*width;
  for (ii = imin; ii <= imax; ii++) {
    for (jj = jmin; jj <= jmax; jj++) {
      ww = w*kernel[(ii + size - i)*(2*size + 1) + (jj + size - j)];
      p = som + (ii*width + (jj + jmv) % width)*bpp;
      for (ch = 0; ch < 3; ch++) {
        p[ch] += ww*vec[ch];
        p[ch] /= (1 + ww);
      }
    }
  }

  if (!use_somtree)
    return;

  if (size >= width/2 - 1) {
    /* horizontal stripe */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          0, imin, width, imax+1);
  }
  else if (jmin >= 0 && jmax < width) {
    /* regular rectangle */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, imin, jmax+1, imax+1);
  }
  else if (jmin < 0) {
    /* rectangle overflowing left */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          0, imin, jmax+1, imax+1);
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin + width, imin, width, imax+1);
  }
  else {
    /* rectangle overflowing right */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          0, imin, jmax - width + 1, imax+1);
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, imin, width, imax+1);
  }
}

void
update_map_v_cylinder(gint width, gint height, gdouble *som,
                      gint m, gdouble w, gdouble *vec,
                      gint size, gdouble *kernel,
                      gint depth, TreeLevel **bbtree,
                      gboolean use_somtree)
{
  gint i, j, ii, jj;
  gint imin, imax, jmin, jmax, imv;
  gint ch;
  gdouble ww;
  gdouble *p;

  i = m/width;
  j = m % width;
  imin = i - size;
  imax = i + size;
  jmin = MAX(j - size, 0);
  jmax = MIN(j + size, width-1);
  imv = (1 + size/height)*height;
  for (ii = imin; ii <= imax; ii++) {
    for (jj = jmin; jj <= jmax; jj++) {
      ww = w*kernel[(ii + size - i)*(2*size + 1) + (jj + size - j)];
      p = som + ((ii + imv) % height * width + jj)*bpp;
      for (ch = 0; ch < 3; ch++) {
        p[ch] += ww*vec[ch];
        p[ch] /= (1 + ww);
      }
    }
  }

  if (!use_somtree)
    return;

  if (size >= height/2 - 1) {
    /* vertical stripe */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, 0, jmax+1, height);
  }
  else if (imin >= 0 && imax < height) {
    /* regular rectangle */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, imin, jmax+1, imax+1);
  }
  else if (imin < 0) {
    /* rectangle overflowing up */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, 0, jmax+1, imax+1);
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, imin + height, jmax+1, height);
  }
  else {
    /* rectangle overflowing down */
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, 0, jmax+1, imax - height + 1);
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          jmin, imin, jmax+1, height);
  }
}

void
update_map_torus(gint width, gint height, gdouble *som,
                 gint m, gdouble w, gdouble *vec,
                 gint size, gdouble *kernel,
                 gint depth, TreeLevel **bbtree,
                 gboolean use_somtree)
{
  gint i, j, ii, jj;
  gint imin, imax, jmin, jmax, imv, jmv;
  gint ch;
  gdouble ww;
  gdouble *p;

  i = m/width;
  j = m % width;
  imin = i - size;
  imax = i + size;
  imv = (1 + size/height)*height;
  jmin = j - size;
  jmax = j + size;
  jmv = (1 + size/width)*width;
  for (ii = imin; ii <= imax; ii++) {
    for (jj = jmin; jj <= jmax; jj++) {
      ww = w*kernel[(ii + size - i)*(2*size + 1) + (jj + size - j)];
      p = som + ((ii + imv) % height * width + (jj + jmv) % width)*bpp;
      for (ch = 0; ch < 3; ch++) {
        p[ch] += ww*vec[ch];
        p[ch] /= (1 + ww);
      }
    }
  }

  if (!use_somtree)
    return;

  if (size >= width/2 - 1) {
    if (size >= height/2 - 1) {
      /* full rectangle */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, width, height);
    }
    else if (imin >= 0 && imax < height) {
      /* horizontal stripe */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin, width, imax+1);
    }
    else if (imin < 0) {
      /* horizontal stripe overflowing up */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, width, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin + height, width, height);
    }
    else {
      /* horizontal stripe overflowing down */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, width, imax - height + 1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin, width, height);
    }
  }
  else if (jmin >= 0 && jmax < width) {
    if (size >= height/2 - 1) {
      /* vertical stripe */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, 0, jmax+1, height);
    }
    else if (imin >= 0 && imax < height) {
      /* regular rectangle */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, imin, jmax+1, imax+1);
    }
    else if (imin < 0) {
      /* rectangle overflowing up */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, 0, jmax+1, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, imin + height, jmax+1, height);
    }
    else {
      /* rectangle overflowing down */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, 0, jmax+1, imax - height + 1);
      update_bounding_boxes(jmax+1, height, bpp, som, depth, bbtree,
                            jmin, imin, jmax+1, height);
    }
  }
  else if (jmin < 0) {
    if (size >= height/2 - 1) {
      /* vertical stripe overflowing left */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, jmax+1, height);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin + width, 0, width, height);
    }
    else if (imin >= 0 && imax < height) {
      /* rectangle overflowing left */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin, jmax+1, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin + width, imin, width, imax+1);
    }
    else if (imin < 0) {
      /* rectangle overflowing up and left */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, jmax+1, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin + height, jmax+1, height);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin + width, 0, width, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin + width, imin + height, width, height);
    }
    else {
      /* rectangle overflowing down and left */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, jmax+1, imax - height + 1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin, jmax+1, height);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin + width, 0, width, imax - height + 1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin + width, imin, width, height);
    }
  }
  else {
    if (size >= height/2 - 1) {
      /* vertical stripe overflowing right */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, jmax - width + 1, height);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, 0, width, height);
    }
    else if (imin >= 0 && imax < height) {
      /* rectangle overflowing right */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin, jmax - width + 1, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, imin, width, imax+1);
    }
    else if (imin < 0) {
      /* rectangle overflowing up and right */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, jmax - width + 1, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin + height, jmax - width + 1, height);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, 0, width, imax+1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, imin + height, width, height);
    }
    else {
      /* rectangle overflowing down and right */
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, 0, jmax - width + 1, imax - height + 1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            0, imin, jmax - width + 1, height);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, 0, width, imax - height + 1);
      update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                            jmin, imin, width, height);
    }
  }
}

