/*
 * @(#) $Id: selfsom.c,v 1.26 2004/08/02 12:14:29 yeti Exp $
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2004 Yeti (David Necas) <yeti@physics.muni.cz>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
   /* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "libgimp/stdplugins-intl.h"
#include "somtree.h"
#include "mapupdate.h"

#define PROCEDURE_NAME   "selfsom"
#define DATA_KEY_VALS    "plug_in_selfsom"

#define ENTRY_WIDTH 75

/* Don't change, though it's ugly */
#define SCALE_WIDTH 201

enum {
  SELFSOM_RESPONSE_RESET = 1,
  SELFSOM_RESPONSE_DEFAULTS,
  SELFSOM_RESPONSE_ABOUT
};

enum {
  SELFSOM_KERNEL_SQUARE,
  SELFSOM_KERNEL_CIRCLE,
  SELFSOM_KERNEL_GAUSSIAN
};

enum {
  SELFSOM_COOLING_EXPONENTIAL,
  SELFSOM_COOLING_LINEAR,
  SELFSOM_COOLING_CAUCHY
};

enum {
  SELFSOM_TOPOLOGY_SQUARE,
  SELFSOM_TOPOLOGY_H_CYLINDER,
  SELFSOM_TOPOLOGY_V_CYLINDER,
  SELFSOM_TOPOLOGY_TORUS
};

typedef struct
{
  gint     training_set;
  gint     kernel_type;
  gint     topology;
  gint     size_cooling;
  gint     weight_cooling;
  gint     iterations;
  gint     start_size;
  gint     end_size;
  gdouble  start_weight;
  gdouble  end_weight;
  gboolean new_img;
  gboolean conserve_mem;
} SelfSOMValues;

/* local prototypes */
static void query                    (void);
static void run                      (const gchar *name,
                                      gint nparams,
                                      const GimpParam *param,
                                      gint *nreturn_vals,
                                      GimpParam **return_vals);
static gint selfsom_dialog           (GimpDrawable *drawable);
static void selfsom                  (GimpDrawable *drawable);
static gint training_set_constrain   (gint32 image_id,
                                      gint32 drawable_id,
                                      gpointer data);
static void training_set_callback    (gint32 id,
                                      gpointer data);
static void size_changed_cb          (gpointer data);
static void weight_changed_cb        (gpointer data);
static void enum_value_changed       (GtkWidget *widget,
                                      guint *value);
static void selfsom_refresh_controls (const SelfSOMValues *new_svals);
static void selfsom_relax            (gpointer *data);

/***** Local vars *****/

typedef struct
{
  GtkWidget *training_set;
  GtkWidget *kernel_type;
  GtkWidget *topology;
  GtkWidget *size_cooling;
  GtkWidget *weight_cooling;
  GtkObject *iterations;
  GtkObject *start_size;
  GtkObject *end_size;
  GtkObject *start_weight;
  GtkObject *end_weight;
  GtkWidget *new_img;
  GtkWidget *conserve_mem;
} SelfSOMControls;

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

static const SelfSOMValues svals_defaults =
{
  -1,                           /* training_set */
  SELFSOM_KERNEL_SQUARE,        /* kernel_type */
  SELFSOM_TOPOLOGY_SQUARE,      /* topology */
  SELFSOM_COOLING_EXPONENTIAL,  /* size_cooling */
  SELFSOM_COOLING_LINEAR,       /* weight_cooling */
  10000,                        /* iterations */
  100,                          /* start_size */
  1,                            /* end_size */
  0.4,                          /* start_weight */
  0.04,                         /* end_weight */
  FALSE,                        /* new_img */
  FALSE,                        /* conserve_mem */
};

/* the order must match the corresponding enum! */
static const MapUpdateFunc map_update_functions[] = {
  &update_map_square,
  &update_map_h_cylinder,
  &update_map_v_cylinder,
  &update_map_torus,
};

static SelfSOMValues svals; /* = svals_defaults; */
static SelfSOMValues svals_old;
static gboolean use_somtree = TRUE;

/* interface to the renderer */
static SelfSOMControls sctrl;

/***** Functions *****/

MAIN()

static void
query(void)
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32,    "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE,    "image", "Input image (unused)" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
    { GIMP_PDB_DRAWABLE, "training_set", "Drawable used as training set" },
    { GIMP_PDB_INT32,    "kernel_type", "Kernel type" },
    { GIMP_PDB_INT32,    "topology", "Map topology" },
    { GIMP_PDB_INT32,    "size_cooling", "Size cooling schedule" },
    { GIMP_PDB_INT32,    "weight_cooling", "Size cooling schedule" },
    { GIMP_PDB_INT32,    "iterations", "Number of iterations" },
    { GIMP_PDB_INT32,    "start_size", "Starting kernel size" },
    { GIMP_PDB_INT32,    "end_size", "Final kernel size" },
    { GIMP_PDB_FLOAT,    "start_weight", "Starting weight" },
    { GIMP_PDB_FLOAT,    "end_weight", "Final weight" },
    { GIMP_PDB_INT32,    "new_img", "Create new image?" },
    { GIMP_PDB_INT32,    "conserve_mem", "Conserve memory? (slow)" },
  };

  gimp_install_procedure(DATA_KEY_VALS,
                         "Turn an image to a Self-Organized Map",
                         "This plug-in turns an image to a Self-Organized "
                           "Map using another image as training set.",
                         "Yeti",
                         "Yeti",
                         "Jun 2004",
                         N_("<Image>/Filters/Colors/Self-SOM..."),
                         "RGB",
                         GIMP_PLUGIN,
                         G_N_ELEMENTS(args), 0,
                         args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam  *param,
    gint *nreturn_vals,
    GimpParam  **return_vals)
{
  static GimpParam values[1];
  GimpDrawable *drawable;
  GimpRunMode run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;

  run_mode = param[0].data.d_int32;

  /*  Initialize i18n support  */
	gchar *LOCALEDIR = _glib_get_locale_dir (); //samj
  bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
  bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
  textdomain(GETTEXT_PACKAGE);

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;

  /*  Get the specified drawable  */
  drawable = gimp_drawable_get(param[2].data.d_drawable);

  svals = svals_defaults;
  switch (run_mode) {
    case GIMP_RUN_INTERACTIVE:
      /*  Possibly retrieve data  */
      gimp_get_data(DATA_KEY_VALS, &svals);

      /*  First acquire information with a dialog  */
      svals_old = svals;

      if (!selfsom_dialog(drawable)) {
        gimp_drawable_detach(drawable);
        return;
      }
      break;

    case GIMP_RUN_NONINTERACTIVE:
      /*  Make sure all the arguments are there!  */
      if (nparams != 15)
        status = GIMP_PDB_CALLING_ERROR;
      else {
        svals.training_set = (gint)param[3].data.d_int32;
        svals.kernel_type = (gint)param[4].data.d_int32;
        svals.topology = (gint)param[5].data.d_int32;
        svals.size_cooling = (gint)param[6].data.d_int32;
        svals.weight_cooling = (gint)param[7].data.d_int32;
        svals.iterations = (gint)param[8].data.d_int32;
        svals.start_size = (gint)param[9].data.d_int32;
        svals.end_size = (gint)param[10].data.d_int32;
        svals.start_weight = (gdouble)param[11].data.d_float;
        svals.end_weight = (gdouble)param[12].data.d_float;
        svals.new_img = (gint)param[13].data.d_int32;
        svals.conserve_mem = (gint)param[14].data.d_int32;
      }
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      gimp_get_data(DATA_KEY_VALS, &svals);
      break;

    default:
      break;
  }

  if (status == GIMP_PDB_SUCCESS) {
    /*  Make sure that the drawable is gray or RGB color  */
    if (gimp_drawable_is_rgb(drawable->drawable_id)
        || gimp_drawable_is_gray(drawable->drawable_id)) {
      gimp_progress_init(_("Self-SOM..."));

      selfsom(drawable);

      if (run_mode != GIMP_RUN_NONINTERACTIVE)
        gimp_displays_flush();

      /*  Store data  */
      if (run_mode == GIMP_RUN_INTERACTIVE)
        gimp_set_data(DATA_KEY_VALS, &svals, sizeof(SelfSOMValues));
    }
    else
      status = GIMP_PDB_EXECUTION_ERROR;
  }

  values[0].data.d_status = status;
  gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline void
yeti_table_attach_to_row(GtkWidget *table,
                         GtkWidget *child,
                         const gint row)
{
  gtk_table_attach(GTK_TABLE(table), child, 0, GTK_TABLE(table)->ncols,
                   row, row+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline GtkWidget*
yeti_frame_new_in_box(const gchar *name,
                      GtkWidget *box,
                      const gboolean expand,
                      const gboolean fill)
{
  GtkWidget *frame;

  frame = gtk_frame_new(name);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start(GTK_BOX(box), frame, expand, fill, 0);
  gtk_widget_show(frame);

  return frame;
}

static inline GtkWidget*
yeti_table_new_in_frame(const gint cols,
                        const gint rows,
                        GtkWidget *frame)
{
  GtkWidget *table;
  GtkWidget *align;

  align = gtk_alignment_new(0.5, 0.0, 1.0, 0.0);
  gtk_container_set_border_width(GTK_CONTAINER(align), 0);
  gtk_container_add(GTK_CONTAINER(frame), align);
  gtk_widget_show(align);

  table = gtk_table_new(cols, rows, FALSE);
  gtk_table_set_col_spacings(GTK_TABLE(table), 4);
  gtk_table_set_row_spacings(GTK_TABLE(table), 2);
  gtk_container_set_border_width(GTK_CONTAINER(table), 4);
  gtk_container_add(GTK_CONTAINER(align), table);
  gtk_widget_show(table);

  return table;
}

static inline GtkWidget*
yeti_check_button_new_with_label(const gchar *name,
                                 gint *value,
                                 GCallback cb,
                                 gpointer data)
{
  GtkWidget *check;

  check = gtk_check_button_new_with_mnemonic(name);
  g_signal_connect(G_OBJECT(check), "toggled",
                   G_CALLBACK(gimp_toggle_button_update), value);
  g_signal_connect_swapped(G_OBJECT(check), "toggled", cb, data);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), *value);
  gtk_widget_show(check);

  return check;
}

static inline GtkObject*
yeti_scale_entry_new_double(const gchar *name,
                            GtkWidget *table,
                            const gint row,
                            double *value,
                            const double min,
                            const double max,
                            const double step,
                            const double page,
                            const gint digits,
                            GCallback cb,
                            gpointer data)
{
  GtkObject *adj;

  adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                             *value, min, max, step, page, digits,
                             TRUE, 0, 0, NULL, NULL);
  g_signal_connect(G_OBJECT(adj), "value_changed",
                   G_CALLBACK(gimp_double_adjustment_update),
                   value);
  g_signal_connect_swapped(G_OBJECT(adj), "value_changed", cb, data);

  return adj;
}

static inline GtkObject*
yeti_scale_entry_new_int(const gchar *name,
                         GtkWidget *table,
                         const gint row,
                         gint *value,
                         const gint min,
                         const gint max,
                         const gint step,
                         const gint page,
                         GCallback cb,
                         gpointer data)
{
  GtkObject *adj;

  adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                             *value, min, max, step, page, 0,
                             TRUE, 0, 0, NULL, NULL);
  g_signal_connect(G_OBJECT(adj), "value_changed",
                   G_CALLBACK(gimp_int_adjustment_update),
                   value);
  g_signal_connect_swapped(G_OBJECT(adj), "value_changed", cb, data);

  return adj;
}

static inline void
yeti_progress_update(gint p, gint max)
{
  double r;

  r = (double)p/max;
  gimp_progress_update(r);
}
/* }}} */

static gint
selfsom_dialog(GimpDrawable *drawable)
{
  GtkWidget *dlg;
  GtkWidget *main_vbox;
  GtkWidget *frame;
  GtkWidget *table;
  GtkWidget *menu;
  gint run, row;

  gimp_ui_init(PROCEDURE_NAME, TRUE);

  dlg = gimp_dialog_new(_("Self-SOM"), PROCEDURE_NAME,
                        NULL, 0,
                        gimp_standard_help_func, "filters/selfsom.html",

                        _("Defaults"), SELFSOM_RESPONSE_DEFAULTS,
                        GIMP_STOCK_RESET, SELFSOM_RESPONSE_RESET,
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                        GTK_STOCK_OK, GTK_RESPONSE_OK,
                        NULL);

  main_vbox = gtk_vbox_new(FALSE, 4);
  gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 6);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), main_vbox, TRUE, TRUE, 0);
  gtk_widget_show(main_vbox);

  /* training_set settings  */
  frame = yeti_frame_new_in_box(_("Training"), main_vbox, TRUE, TRUE);
  table = yeti_table_new_in_frame(3, 4, frame);
  row = 0;

  /* training_set source */
  if (!training_set_constrain(0, svals.training_set, drawable))
    svals.training_set = drawable->drawable_id;
  sctrl.training_set = gtk_option_menu_new();
  menu = gimp_drawable_menu_new(training_set_constrain, training_set_callback,
                                drawable, svals.training_set);
  gtk_option_menu_set_menu(GTK_OPTION_MENU(sctrl.training_set), menu);
  gtk_widget_show(sctrl.training_set);
  gimp_table_attach_aligned(GTK_TABLE(table), 0, row++, _("_Training set:"),
                            1.0, 0.5, sctrl.training_set, 3, TRUE);

  sctrl.kernel_type
    = gimp_int_option_menu_new(FALSE, G_CALLBACK(enum_value_changed),
                               &svals.kernel_type, svals.kernel_type,
                               _("Square"), SELFSOM_KERNEL_SQUARE, NULL,
                               _("Circle"), SELFSOM_KERNEL_CIRCLE, NULL,
                               _("Gaussian"), SELFSOM_KERNEL_GAUSSIAN, NULL,
                               NULL);
  gimp_table_attach_aligned(GTK_TABLE(table), 0, row++,
                            _("_Kernel type:"), 1.0, 0.5,
                            sctrl.kernel_type, 2, TRUE);

  sctrl.iterations
    = yeti_scale_entry_new_int(_("_Iterations:"),
                               table, row++, &svals.iterations,
                               1, 200000, 100, 1000,
                               G_CALLBACK(selfsom_relax), NULL);

  sctrl.topology
    = gimp_int_option_menu_new(FALSE, G_CALLBACK(enum_value_changed),
                               &svals.topology, svals.topology,
                               _("Square"),
                               SELFSOM_TOPOLOGY_SQUARE, NULL,
                               _("Horizontal cylinder"),
                               SELFSOM_TOPOLOGY_H_CYLINDER, NULL,
                               _("Vertical cylinder"),
                               SELFSOM_TOPOLOGY_V_CYLINDER, NULL,
                               _("Torus"),
                               SELFSOM_TOPOLOGY_TORUS, NULL,
                               NULL);
  gimp_table_attach_aligned(GTK_TABLE(table), 0, row++,
                            _("Map _topology:"), 1.0, 0.5,
                            sctrl.topology, 2, TRUE);

  frame = yeti_frame_new_in_box(_("Size Cooling"), main_vbox, TRUE, TRUE);
  table = yeti_table_new_in_frame(3, 3, frame);
  row = 0;

  sctrl.size_cooling
    = gimp_int_option_menu_new(FALSE, G_CALLBACK(enum_value_changed),
                               &svals.size_cooling, svals.size_cooling,
                               _("Exponential"),
                               SELFSOM_COOLING_EXPONENTIAL, NULL,
                               _("Linear"),
                               SELFSOM_COOLING_LINEAR, NULL,
                               _("Cauchy"),
                               SELFSOM_COOLING_CAUCHY, NULL,
                               NULL);
  gimp_table_attach_aligned(GTK_TABLE(table), 0, row++,
                            _("S_chedule:"), 1.0, 0.5,
                            sctrl.size_cooling, 2, TRUE);

  sctrl.start_size
    = yeti_scale_entry_new_int(_("_Start size:"),
                               table, row++, &svals.start_size,
                               1, 256, 1, 10,
                               G_CALLBACK(size_changed_cb), GINT_TO_POINTER(0));
  sctrl.end_size
    = yeti_scale_entry_new_int(_("_End size:"),
                               table, row++, &svals.end_size,
                               1, 256, 1, 10,
                               G_CALLBACK(size_changed_cb), GINT_TO_POINTER(1));

  frame = yeti_frame_new_in_box(_("Weight Cooling"), main_vbox, TRUE, TRUE);
  table = yeti_table_new_in_frame(3, 3, frame);
  row = 0;

  sctrl.weight_cooling
    = gimp_int_option_menu_new(FALSE, G_CALLBACK(enum_value_changed),
                               &svals.weight_cooling, svals.weight_cooling,
                               _("Exponential"),
                               SELFSOM_COOLING_EXPONENTIAL, NULL,
                               _("Linear"),
                               SELFSOM_COOLING_LINEAR, NULL,
                               _("Cauchy"),
                               SELFSOM_COOLING_CAUCHY, NULL,
                               NULL);
  gimp_table_attach_aligned(GTK_TABLE(table), 0, row++,
                            _("Sc_hedule:"), 1.0, 0.5,
                            sctrl.weight_cooling, 2, TRUE);

  sctrl.start_weight
    = yeti_scale_entry_new_double(_("Start _weight:"),
                                  table, row++,
                                  &svals.start_weight,
                                  0.001, 0.999, 0.001, 0.1, 3,
                                  G_CALLBACK(weight_changed_cb),
                                  GINT_TO_POINTER(0));
  sctrl.end_weight
    = yeti_scale_entry_new_double(_("End weig_ht:"),
                                  table, row++,
                                  &svals.end_weight,
                                  0.001, 0.999, 0.001, 0.1, 3,
                                  G_CALLBACK(weight_changed_cb),
                                  GINT_TO_POINTER(1));

  frame = yeti_frame_new_in_box(_("Options"), main_vbox, TRUE, TRUE);
  table = yeti_table_new_in_frame(3, 2, frame);
  row = 0;

  sctrl.new_img = yeti_check_button_new_with_label(_("Create _new image"),
                                                   &svals.new_img,
                                                   G_CALLBACK(selfsom_relax),
                                                   NULL);
  yeti_table_attach_to_row(table, sctrl.new_img, row++);

  sctrl.conserve_mem = yeti_check_button_new_with_label(_("Conserve _memory "
                                                          "sacrificing speed"),
                                                &svals.conserve_mem,
                                                G_CALLBACK(selfsom_relax),
                                                NULL);
  yeti_table_attach_to_row(table, sctrl.conserve_mem, row++);

  gtk_widget_show(dlg);

  do {
    run = gimp_dialog_run(GIMP_DIALOG(dlg));
    if (run == SELFSOM_RESPONSE_RESET)
      selfsom_refresh_controls(&svals_old);
    else if (run == SELFSOM_RESPONSE_DEFAULTS)
      selfsom_refresh_controls(&svals_defaults);
    else if (run == GTK_RESPONSE_OK) {
      g_assert(svals.training_set != -1);
      break;
    }
    else
      break;
  } while (TRUE);

  gtk_widget_destroy(dlg);
  while (gtk_events_pending())
    gtk_main_iteration_do(FALSE);

  return run == GTK_RESPONSE_OK;
}

static void
size_changed_cb(gpointer data)
{
  if (svals.end_size <= svals.start_size)
    return;

  if (data) {
    svals.start_size = svals.end_size;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.start_size),
                             svals.start_size);
  }
  else {
    svals.end_size = svals.start_size;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.end_size),
                             svals.end_size);
  }
}

static void
weight_changed_cb(gpointer data)
{
  if (svals.end_weight <= svals.start_weight)
    return;

  if (data) {
    svals.start_weight = svals.end_weight;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.start_weight),
                             svals.start_weight);
  }
  else {
    svals.end_weight = svals.start_weight;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.end_weight),
                             svals.end_weight);
  }
}

static void
enum_value_changed(GtkWidget *widget,
                   guint *value)
{
  guint v;

  v = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(widget), "gimp-item-data"));
  *value = v;
}


static void
selfsom_relax(gpointer *data)
{
  /* do nothing, but stay ready ;-) */
  /* FIXME */
}

/* training set must be RGB, like the image itself */
static gint
training_set_constrain(gint32 image_id, gint32 drawable_id, gpointer data)
{
  if (drawable_id == -1)
    return TRUE;

  return gimp_drawable_is_rgb(drawable_id)
         && !gimp_drawable_has_alpha(drawable_id);
}

static void
training_set_callback(gint32 id, gpointer data)
{
  svals.training_set = id;
}

/* though this is obscure, we use it to reset/revert all values _and_ refresh
   the dialog controls */
static void
selfsom_refresh_controls(const SelfSOMValues *new_svals)
{
  /* we cannot reset some field and don't want to reset some others, so copy
     them one by one  */
  svals.iterations = new_svals->iterations;
  gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.iterations),
                           svals.iterations);

  svals.start_size = new_svals->start_size;
  gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.start_size),
                           svals.start_size);

  svals.end_size = new_svals->end_size;
  gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.end_size),
                           svals.end_size);

  svals.start_weight = new_svals->start_weight;
  gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.start_weight),
                           svals.start_weight);

  svals.end_weight = new_svals->end_weight;
  gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(sctrl.end_weight),
                           svals.end_weight);

  svals.kernel_type = new_svals->kernel_type;
  gimp_option_menu_set_history(GTK_OPTION_MENU(sctrl.kernel_type),
                               GUINT_TO_POINTER(svals.kernel_type));

  svals.topology = new_svals->topology;
  gimp_option_menu_set_history(GTK_OPTION_MENU(sctrl.topology),
                               GUINT_TO_POINTER(svals.topology));

  svals.size_cooling = new_svals->size_cooling;
  gimp_option_menu_set_history(GTK_OPTION_MENU(sctrl.size_cooling),
                               GUINT_TO_POINTER(svals.size_cooling));

  svals.weight_cooling = new_svals->weight_cooling;
  gimp_option_menu_set_history(GTK_OPTION_MENU(sctrl.weight_cooling),
                               GUINT_TO_POINTER(svals.weight_cooling));

  svals.new_img = new_svals->new_img;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sctrl.new_img), svals.new_img);

  svals.conserve_mem = new_svals->conserve_mem;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sctrl.conserve_mem),
                               svals.conserve_mem);

  /* don't want to update: training_set
  */
}

static void
compute_kernel(gint kernel_type,
               gint size,
               gdouble weight,
               gdouble *kernel)
{
  gint i, j, n;

  if (size == 0) {
    kernel[0] = weight;
    return;
  }

  n = 2*size + 1;
  switch (kernel_type) {
    case SELFSOM_KERNEL_SQUARE:
    for (i = 0; i < n*n; i++)
      kernel[i] = weight;
    break;

    case SELFSOM_KERNEL_CIRCLE:
    weight *= 4.0/G_PI;
    memset(kernel, 0, n*n*sizeof(gdouble));
    for (i = 0; i < n; i++) {
      gint r = floor(sqrt((2.0*i + 1.0)*size - i*i + 0.25));

      for (j = size - r; j <= size + r; j++)
        kernel[i*n + j] = weight;
    }
    break;

    case SELFSOM_KERNEL_GAUSSIAN:
    weight *= 16.0/G_PI;
    for (i = 0; i < n; i++) {
      gdouble q = exp(-(i - size)*(i - size)*4.0/(size*size))*weight;
      for (j = 0; j < n; j++)
        kernel[i*n + j] = q;
    }
    for (j = 0; j < n; j++) {
      gdouble q = exp(-(j - size)*(j - size)*4.0/(size*size));
      for (i = 0; i < n; i++)
        kernel[i*n + j] *= q;
    }
    break;

    default:
    g_assert_not_reached();
    break;
  }
}

static gdouble
interpolate(gdouble x, guint cooling,
            gdouble start, gdouble end)
{
  switch (cooling) {
    case SELFSOM_COOLING_CAUCHY:
    return start/(x*(start/end - 1.0) + 1.0);
    break;

    case SELFSOM_COOLING_LINEAR:
    return x*(end - start) + start;
    break;

    case SELFSOM_COOLING_EXPONENTIAL:
    return start*exp(log(end/start)*x);
    break;

    default:
    g_assert_not_reached();
    break;
  }

  return -100000;  /* coredump... */
}

static void
selfsom(GimpDrawable *drawable)
{
  GimpDrawable *som_drawable, *tset_drawable;
  gint32 som_lay, tset_lay,
         som_img = sizeof("No, stupid gcc, this can't be used uninitialized");
  GimpPixelRgn image_rgn, som_rgn, tset_rgn;
  GimpPixelFetcher *pfetch;
  guchar *image;
  guchar *tset;
  gdouble *som;
  gdouble *kernel;
  gdouble q, w;
  gint channels, bpp;
  gint width, height;
  gint tset_width, tset_height;
  gint i, n, y;
  gint size, oldsize;
  gpointer pr;
  TreeLevel **bbtree = NULL;
  MapUpdateFunc update_map;
  gint depth;
  GRand *rng;
  gboolean use_pixel_fetcher = FALSE;

  width = gimp_drawable_width(drawable->drawable_id);
  height = gimp_drawable_height(drawable->drawable_id);
  use_somtree = width > 8 && height > 8;  /* FIXME: depends on alpha */
  g_assert(svals.topology >= 0
           && svals.topology < G_N_ELEMENTS(map_update_functions));
  update_map = map_update_functions[svals.topology];

  tset_lay = svals.training_set;
  tset_height = gimp_drawable_height(tset_lay);
  tset_width = gimp_drawable_width(tset_lay);

  /* set up all the tile and pixel region stuff */
  gimp_tile_cache_ntiles(1);

  channels = bpp = gimp_drawable_bpp(drawable->drawable_id);
  if (gimp_drawable_has_alpha(drawable->drawable_id))
    channels--;

  som = g_new(gdouble, bpp*width*height);
  image = g_new(guchar, bpp*width*height);
  if (!use_pixel_fetcher)
    tset = g_new(guchar, bpp*tset_width*tset_height);

  tset_drawable = gimp_drawable_get(tset_lay);
  gimp_pixel_rgn_init(&image_rgn, drawable, 0, 0,
                      width, height, FALSE, FALSE);
  if (!use_pixel_fetcher)
    gimp_pixel_rgn_init(&tset_rgn, tset_drawable, 0, 0,
                        tset_width, tset_height, FALSE, FALSE);

  for (pr = gimp_pixel_rgns_register(1, &image_rgn);
       pr != NULL;
       pr = gimp_pixel_rgns_process(pr)) {
    for (y = image_rgn.y; y < image_rgn.y + image_rgn.h; y++)
      memcpy(image + y*width*bpp + image_rgn.x*bpp,
             image_rgn.data + (y - image_rgn.y)*image_rgn.rowstride,
             image_rgn.w*bpp);
  }

  for (i = 0; i < width*height*bpp; i++)
    som[i] = image[i];

  if (use_somtree) {
    bbtree = allocate_tree(width, height, bpp, &depth);
    update_bounding_boxes(width, height, bpp, som, depth, bbtree,
                          0, 0, width, height);
  }

  if (!use_pixel_fetcher) {
    for (pr = gimp_pixel_rgns_register(1, &tset_rgn);
         pr != NULL;
         pr = gimp_pixel_rgns_process(pr)) {
      for (y = tset_rgn.y; y < tset_rgn.y + tset_rgn.h; y++)
        memcpy(tset + y*tset_width*bpp + tset_rgn.x*bpp,
               tset_rgn.data + (y - tset_rgn.y)*tset_rgn.rowstride,
               tset_rgn.w*bpp);
    }
  }

  oldsize = -1;
  rng = g_rand_new();
  kernel = g_new(gdouble, (2*svals.start_size - 1)*(2*svals.start_size - 1));
  if (use_pixel_fetcher)
    pfetch = gimp_pixel_fetcher_new(tset_drawable, FALSE);
  for (n = 0; n < svals.iterations; n++) {
    gint m, ch;
    guchar pixel[3];
    gdouble vec[3];

    q = (gdouble)n/svals.iterations;
    /* -1 because 1 is minimum in GUI 1 but we compute the size as 2*s+1 */
    size = interpolate(q, svals.size_cooling,
                       svals.start_size, svals.end_size) - 1;
    size = MIN(size, svals.start_size-1);    /* paranoia */
    w = interpolate(q, svals.weight_cooling,
                    svals.start_weight, svals.end_weight);
    w = w/(1.0 - w);
    if (size != oldsize) {
      compute_kernel(svals.kernel_type, size, 1.0, kernel);
      oldsize = size;
    }
    i = g_rand_int_range(rng, 0, tset_width*tset_height);
    if (use_pixel_fetcher) {
      gimp_pixel_fetcher_get_pixel(pfetch, i % tset_width, i/tset_width, pixel);
      for (ch = 0; ch < 3; ch++)
        vec[ch] = pixel[ch];
    }
    else {
      for (ch = 0; ch < 3; ch++)
        vec[ch] = tset[i*bpp + ch];
    }

    if (use_somtree)
      m = find_best_match(width, height, bpp, som, depth, bbtree, vec);
    else
      m = find_best_match_slow(width, height, bpp, som, vec);

    update_map(width, height, som, m, w, vec, size, kernel,
               depth, bbtree, use_somtree);

    if (!(n % 100)) {
      gimp_progress_update(n/(gdouble)svals.iterations);
      /*fprintf(stderr, "%d %d\n", n, size);*/
    }
  }

  if (use_pixel_fetcher)
    gimp_pixel_fetcher_destroy(pfetch);
  else
    g_free(tset);
  g_free(kernel);
  g_rand_free(rng);

  for (i = 0; i < width*height*bpp; i++)
    image[i] = (gint)floor(som[i] + 0.4999);

  g_free(som);

  /* eventually create a new image */
  if (svals.new_img) {
    som_img = gimp_image_new(width, height, GIMP_RGB);
    som_lay = gimp_layer_new(som_img, "SOM",
                            width, height,
                            GIMP_RGB_IMAGE, 100, GIMP_NORMAL_MODE);
    gimp_image_add_layer(som_img, som_lay, 0);
    som_drawable = gimp_drawable_get(som_lay);
  }
  else
    som_drawable = drawable;

  gimp_pixel_rgn_init(&som_rgn, som_drawable, 0, 0,
                      width, height, TRUE, TRUE);

  for (pr = gimp_pixel_rgns_register(1, &som_rgn);
       pr != NULL;
       pr = gimp_pixel_rgns_process(pr)) {
    for (y = som_rgn.y; y < som_rgn.y + som_rgn.h; y++)
      memcpy(som_rgn.data + (y - som_rgn.y)*som_rgn.rowstride,
             image + y*width*bpp + som_rgn.x*bpp,
             som_rgn.w*bpp);
  }

  g_free(image);

  gimp_drawable_flush(som_drawable);
  gimp_drawable_merge_shadow(som_drawable->drawable_id, TRUE);
  gimp_drawable_update(som_drawable->drawable_id,
                       0, 0, width, height);
  if (svals.new_img) {
    gimp_drawable_detach(som_drawable);
    gimp_display_new(som_img);
  }
}

