
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "somtree.h"

/* Tree fan-out; the optimum is usually between 3 and 4, but 4 is a `nicer'
 * number and eats somwehat less memory */
#define ALPHA 4

/* Fixed len, we only support RGB anyway and it helps compiler to unroll
 * lopps over len.
 * For general use replace LEN with len and remove the asserts LEN == len */
#define LEN 3

TreeLevel**
allocate_tree(gint width, gint height, gint len,
              gint *depth)
{
  TreeLevel **bbtree;
  TreeLevel *tree;
  gint dp, i, x;

  x = MAX(width, height);
  dp = 0;
  for (i = 1; i < x; i *= ALPHA)
    dp++;

  dp--;
  bbtree = g_new(TreeLevel*, dp);
  dp--;

  for (i = dp; i >= 0; i--) {
    bbtree[i] = tree = g_new(TreeLevel, 1);
    tree->width = width = (width + ALPHA-1)/ALPHA;
    tree->height = height = (height + ALPHA-1)/ALPHA;
    tree->min = g_new(gdouble, width*height*len);
    tree->max = g_new(gdouble, width*height*len);
    /*fprintf(stderr, "Tree level %d is %d x %d\n", i, width, height);*/
  }
  g_assert(bbtree[0]->width <= ALPHA && bbtree[0]->height <= ALPHA);

  *depth = dp;
  return bbtree;
}

gdouble*
tree_statistics(gint width, gint height, gint len, gint depth,
                TreeLevel **bbtree)
{
  TreeLevel *tree;
  gdouble *s;
  gdouble z;
  gint widthA, heightA;
  gint i, dp, ch, n;
  gdouble *pmin, *pmax;

  /*fprintf(stderr, "Average bounding box sizes\n");*/
  s = g_new(gdouble, len*(depth + 1));
  for (dp = depth; dp >= 0; dp--) {
    tree = bbtree[dp];
    widthA = tree->width;
    heightA = tree->height;
    memset(s, 0, sizeof(gdouble)*len);
    n = widthA*heightA;
    for (i = 0; i < n; i++) {
      pmin = tree->min + i*len;
      pmax = tree->max + i*len;
      for (ch = 0; ch < len; ch++) {
        z = pmax[ch] - pmin[ch];
        s[dp*len + ch] += z*z;
      }
    }
    /*
    fprintf(stderr, "  %d: %g, %g, %g\n",
            dp, sqrt(s[0]/n), sqrt(s[1]/n), sqrt(s[2]/n));
            */
  }
  for (i = 0; i < len*(depth + 1); i++)
    s[i] = sqrt(s[i]);

  return s;
}

void
update_bounding_boxes(gint width, gint height, gint len,
                      gdouble *som,
                      gint depth,
                      TreeLevel **bbtree,
                      gint ulcol, gint ulrow,
                      gint brcol, gint brrow)
{
  gint i, j, ii, jj, ch, dp;

  gdouble *p, *pmin, *pmax;
  TreeLevel *tree, *base;
  gdouble *min, *max, z;

  g_assert(len == LEN);
  /*fprintf(stderr, "Updating region (%d,%d) x (%d,%d)\n",
          ulcol, brcol, ulrow, brrow);*/
  /* som to leafs */
  ulrow = ulrow/ALPHA;
  ulcol = ulcol/ALPHA;
  brrow = (brrow + ALPHA-1)/ALPHA;
  brcol = (brcol + ALPHA-1)/ALPHA;
  tree = bbtree[depth];
  int widthA = tree->width;
  int heightA = tree->height;
  /*fprintf(stderr, "  At depth %d: updating (%d,%d) x (%d,%d)\n",
          depth, ulcol, brcol, ulrow, brrow);*/
  for (i = ulrow; i < brrow; i++) {
    for (j = ulcol; j < brcol; j++) {
      min = tree->min + (i*widthA + j)*LEN;
      max = tree->max + (i*widthA + j)*LEN;
      for (ch = 0; ch < LEN; ch++) {
        min[ch] = G_MAXDOUBLE;
        max[ch] = -G_MAXDOUBLE;
      }
      for (ii = 0; ii < MIN(ALPHA, height - ALPHA*i); ii++) {
        for (jj = 0; jj < MIN(ALPHA, width - ALPHA*j); jj++) {
          /*fprintf(stderr, "i = %d, ii = %d, j = %d, jj = %d\n",
                  i, ii, j, jj);*/
          p = som + ((i*ALPHA + ii)*width + (j*ALPHA + jj))*LEN;
          for (ch = 0; ch < LEN; ch++) {
            z = p[ch];
            if (z < min[ch])
              min[ch] = z;
            if (z > max[ch])
              max[ch] = z;
          }
        }
      }
      for (ch = 0; ch < LEN; ch++) {
        g_assert(min[ch] != G_MAXDOUBLE);
        g_assert(max[ch] != -G_MAXDOUBLE);
      }
    }
  }

  /* propagate branches up */
  for (dp = depth-1; dp >= 0; dp--) {
    tree = bbtree[dp];
    base = bbtree[dp+1];
    width = base->width;
    height = base->height;
    widthA = tree->width;
    heightA = tree->height;
    ulrow = ulrow/ALPHA;
    ulcol = ulcol/ALPHA;
    brrow = (brrow + ALPHA-1)/ALPHA;
    brcol = (brcol + ALPHA-1)/ALPHA;
    /*fprintf(stderr, "  At depth %d: updating (%d,%d) x (%d,%d)\n",
            dp, ulcol, brcol, ulrow, brrow);*/
    for (i = ulrow; i < brrow; i++) {
      for (j = ulcol; j < brcol; j++) {
        min = tree->min + (i*widthA + j)*LEN;
        max = tree->max + (i*widthA + j)*LEN;
        for (ch = 0; ch < LEN; ch++) {
          min[ch] = G_MAXDOUBLE;
          max[ch] = -G_MAXDOUBLE;
        }
        for (ii = 0; ii < MIN(ALPHA, height - ALPHA*i); ii++) {
          for (jj = 0; jj < MIN(ALPHA, width - ALPHA*j); jj++) {
            pmin = base->min + ((i*ALPHA + ii)*width + (j*ALPHA + jj))*LEN;
            pmax = base->max + ((i*ALPHA + ii)*width + (j*ALPHA + jj))*LEN;
            for (ch = 0; ch < LEN; ch++) {
              if (pmin[ch] < min[ch])
                min[ch] = pmin[ch];
              if (pmax[ch] > max[ch])
                max[ch] = pmax[ch];
            }
          }
        }
      }
    }
  }
}

gint
find_best_match(gint width, gint height, gint len,
                gdouble *som,
                gint depth,
                TreeLevel **bbtree,
                gdouble *vec)
{
  gint i, j, ii, jj, ch, dp, mi, mj;
  gint widthA, heightA;
  gdouble *p, *pmin, *pmax;
  TreeLevel *tree;
  gdouble z, s, ms;
  gdouble bestmin[len], bestmax[len];
  gboolean updated;
  gint besti;

  /*fprintf(stderr, "Looking for (%g, %g, %g)\n", vec[0], vec[1], vec[2]);*/

  g_assert(len == LEN);
  /* make initial guess to start with relatively small bbox */
  /* tree */
  i = j = 0;
  for (dp = 0; dp <= depth; dp++) {
    tree = bbtree[dp];
    widthA = tree->width;
    heightA = tree->height;
    i *= ALPHA;
    j *= ALPHA;
    ms = G_MAXDOUBLE;
    mi = mj = 0;
    for (ii = 0; ii < MIN(ALPHA, heightA - i); ii++) {
      for (jj = 0; jj < MIN(ALPHA, widthA - j); jj++) {
        pmin = tree->min + ((i + ii)*widthA + (j + jj))*LEN;
        pmax = tree->max + ((i + ii)*widthA + (j + jj))*LEN;
        s = 0.0;
        for (ch = 0; ch < LEN; ch++) {
          z = pmax[ch] - pmin[ch];
          s += z*z;
          z = 2*vec[ch] - (pmin[ch] + pmax[ch]);
          /* the 5 is a measure of conservativness
           * high values find good matches but may miss
           * small values find bad matches with certainity ;-)
           */
          s += 5*z*z;
        }
        if (s < ms) {
          mi = ii;
          mj = jj;
          ms = s;
        }
      }
    }
    i += mi;
    j += mj;
  }

  /* leafs (som itself) */
  i *= ALPHA;
  j *= ALPHA;
  ms = G_MAXDOUBLE;
  mi = mj = 0;
  for (ii = 0; ii < MIN(ALPHA, height - i); ii++) {
    for (jj = 0; jj < MIN(ALPHA, width - j); jj++) {
      p = som + ((i + ii)*width + (j + jj))*LEN;
      s = 0.0;
      for (ch = 0; ch < LEN; ch++) {
        z = vec[ch] - p[ch];
        s += z*z;
      }
      if (s < ms) {
          mi = ii;
          mj = jj;
          ms = s;
      }
    }
  }
  mi += i;
  mj += j;
  besti = mi*width + mj;
  s = 0.0;
  for (ch = 0; ch < LEN; ch++) {
    z = vec[ch] - som[LEN*besti + ch];
    s += z*z;
  }
  s = sqrt(s);
  for (ch = 0; ch < LEN; ch++) {
    bestmin[ch] = vec[ch] - s;
    bestmax[ch] = vec[ch] + s;
  }

  /*
  fprintf(stderr, "Initial guess (%d,%d) = (%g, %g, %g)\n",
          i, j,
          som[(i*width + j)*LEN],
          som[(i*width + j)*LEN + 1],
          som[(i*width + j)*LEN + 2]);
          */

  /* now walk though the tree (depth-first search)
   * until we are sure we found the best match */
  dp = 0;
  tree = bbtree[dp];
  tree->iter = 0;
  i = j = 0;
  while (dp || tree->iter < ALPHA*ALPHA) {
    /* finished a level, move up */
    if (tree->iter == ALPHA*ALPHA) {
      i = i/ALPHA - i/ALPHA % ALPHA;
      j = j/ALPHA - j/ALPHA % ALPHA;
      dp--;
      tree = bbtree[dp];
      tree->iter++;
      continue;
    }

    /* check out-of-map */
    ii = tree->iter/ALPHA;
    jj = tree->iter % ALPHA;
    if (i + ii >= tree->height) {
      tree->iter = ALPHA*ALPHA;
      continue;
    }
    if (j + jj >= tree->width) {
      tree->iter = (tree->iter + ALPHA-1)/ALPHA*ALPHA;
      continue;
    }

    /* does it make sense to move down here? */
    pmin = tree->min + ((i + ii)*tree->width + (j + jj))*LEN;
    pmax = tree->max + ((i + ii)*tree->width + (j + jj))*LEN;
    for (ch = 0; ch < LEN; ch++) {
      /* interval intersection */
      if (bestmax[ch] < pmin[ch] || pmax[ch] < bestmin[ch])
        break;
    }
    if (ch != LEN) {
      tree->iter++;
      continue;
    }

    /* ok, move down, if we can */
    dp++;
    i = ALPHA*(i + ii);
    j = ALPHA*(j + jj);
    if (dp > depth) {
      /* can't move down any more, scan som tile */
      updated = FALSE;
      for (ii = 0; ii < MIN(ALPHA, height - i); ii++) {
        for (jj = 0; jj < MIN(ALPHA, width - j); jj++) {
          p = som + ((i + ii)*width + (j + jj))*LEN;
          s = 0.0;
          for (ch = 0; ch < LEN; ch++) {
            z = p[ch] - vec[ch];
            s += z*z;
          }
          if (s < ms) {
            ms = s;
            mi = i + ii;
            mj = j + jj;
            updated = TRUE;
          }
        }
      }
      if (updated) {
        besti = mi*width + mj;
        s = 0.0;
        for (ch = 0; ch < LEN; ch++) {
          z = vec[ch] - som[LEN*besti + ch];
          s += z*z;
        }
        s = sqrt(s);
        for (ch = 0; ch < LEN; ch++) {
          bestmin[ch] = vec[ch] - s;
          bestmax[ch] = vec[ch] + s;
        }
      }
      i = i/ALPHA - i/ALPHA % ALPHA;
      j = j/ALPHA - j/ALPHA % ALPHA;
      dp--;
      tree->iter++;
    }
    else {
      tree = bbtree[dp];
      tree->iter = 0;
    }
  }
  /*fprintf(stderr, "  (%g, %g, %g)\n",
          som[LEN*besti], som[LEN*besti + 1], som[LEN*besti + 2]);*/

  return besti;
}

gint
find_best_match_slow(gint width, gint height, gint len,
                     gdouble *som,
                     gdouble *vec)
{
  gint ch, m, i;
  gdouble mm, z, s;
  gdouble *p;

  g_assert(len == LEN);

  m = 0;
  mm = G_MAXDOUBLE;

  for (i = 0; i < width*height; i++) {
    p = som + i*LEN;
    s = 0.0;
    for (ch = 0; ch < LEN; ch++) {
      z = p[ch] - vec[ch];
      s += z*z;
    }
    if (s < mm) {
      mm = s;
      m = i;
    }
  }

  return m;
}

