/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "speedline_preview.h"
#include "speedline_radiation.h"
#include "speedline_line.h"

typedef enum _RadiationGuideObj {
	RADIATION_GUIDE_OBJ_NONE = 0,
	RADIATION_GUIDE_OBJ_CENTER,
	RADIATION_GUIDE_OBJ_IN_CIRCUM,
	RADIATION_GUIDE_OBJ_OUT_CIRCUM,
} RadiationGuideObj;

static void path_guide_in_circum (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius);
static void path_guide_out_circum (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius);
static void path_guide_center (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius);
static void get_global_vals (PluginVals *vals, gdouble *cx, gdouble *cy, gdouble *in_radius, gdouble *out_radius);
static RadiationGuideObj radiation_guide_hit_test (cairo_t *cr, GdkWindow *window, PluginVals *vals, gdouble x, gdouble y);


static void
path_guide_in_circum (cairo_t *cr,
					  gdouble cx,
					  gdouble cy,
					  gdouble in_radius,
					  gdouble out_radius)
{
	cairo_arc (cr, cx, cy, in_radius, 0, 2 * G_PI);
}

static void
path_guide_out_circum (cairo_t *cr, 
					   gdouble cx,
					   gdouble cy,
					   gdouble in_radius,
					   gdouble out_radius)
{
	cairo_arc (cr, cx, cy, out_radius, 0, 2 * G_PI);
}

static void
path_guide_center (cairo_t *cr,
				   gdouble cx,
				   gdouble cy,
				   gdouble in_radius,
				   gdouble out_radius)
{
	double size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;
	cairo_arc (cr, cx, cy, preview_local_to_global (size), 0, 2 * G_PI);
}

static void
get_global_vals (PluginVals *vals,
				 gdouble	*cx, 
				 gdouble	*cy, 
				 gdouble	*in_radius, 
				 gdouble	*out_radius)
{
	gdouble resx, resy;
	gimp_image_get_resolution (vals->image_ID, &resx, &resy);
	*cx = gimp_units_to_pixels (vals->radiation_x, vals->area_unit, resx);
	*cy = gimp_units_to_pixels (vals->radiation_y, vals->area_unit, resy);
	*in_radius  = gimp_units_to_pixels (vals->radiation_in_radius, vals->area_unit, resx);
	*out_radius = gimp_units_to_pixels (vals->radiation_out_radius, vals->area_unit, resx);
}

static RadiationGuideObj
radiation_guide_hit_test (cairo_t		*cr,
						  GdkWindow		*window,
						  PluginVals	*vals,
						  gdouble		x,
						  gdouble		y)
{
	gdouble cx, cy, in_radius, out_radius;

	get_global_vals (vals, &cx, &cy, &in_radius, &out_radius);

	cairo_set_source_rgba (cr, 1, 1, 1, 1);
	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH_FOR_HITTEST));
	cairo_device_to_user (cr, &x, &y);

	//check if in circum
	path_guide_in_circum (cr, cx, cy, in_radius, out_radius);
	if (cairo_in_stroke (cr, x, y)) {
		return RADIATION_GUIDE_OBJ_IN_CIRCUM;
	}
	cairo_new_path (cr);
	
	//check if out circum
	path_guide_out_circum (cr, cx, cy, in_radius, out_radius);
	if (cairo_in_stroke (cr, x, y)) {
		return RADIATION_GUIDE_OBJ_OUT_CIRCUM;
	}
	cairo_new_path (cr);

	//check if center
	path_guide_center (cr, cx, cy, in_radius, out_radius);
	if (cairo_in_fill (cr, x, y)) {
		return RADIATION_GUIDE_OBJ_CENTER;
	}
	cairo_new_path (cr);

	return RADIATION_GUIDE_OBJ_NONE;
}

void
radiation_draw_guide (cairo_t		*cr,
					  PluginVals	*vals)
{
	int n, nmax;
	double sx, sy, dx, dy;
	gdouble cx, cy, in_radius, out_radius;

	get_global_vals (vals, &cx, &cy, &in_radius, &out_radius);

	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH));

	cairo_set_source_rgba (cr,
		PREVIEW_GUIDE_COLOR_R,
		PREVIEW_GUIDE_COLOR_G,
		PREVIEW_GUIDE_COLOR_B,
		1);

	//out
	path_guide_out_circum (cr, cx, cy, in_radius, out_radius);
	cairo_stroke (cr);

	//in
	path_guide_in_circum (cr, cx, cy, in_radius, out_radius);
	cairo_stroke (cr);

	//center
	path_guide_center (cr, cx, cy, in_radius, out_radius);
	cairo_fill (cr);

	//iines
	sx = cx;
	sy = cy;
	dx = cx + in_radius;
	dy = cy;
	nmax = PREVIEW_GUIDE_LINE_NUMBER;
	for (n = 0; n < nmax; n++) {
		// X : CenterX + R * Math.cos(Radian)
		// Y : CenterY - R * Math.sin(Radian)
		sx = cx + in_radius * cos ((double)n/nmax * G_PI * 2);
		sy = cy - in_radius * sin ((double)n/nmax * G_PI * 2);
		dx = cx + out_radius * cos ((double)n/nmax * G_PI * 2);
		dy = cy - out_radius * sin ((double)n/nmax * G_PI * 2);
		cairo_move_to (cr, sx, sy);
		cairo_line_to (cr, dx, dy);
	}
	cairo_stroke (cr);
}

void
radiation_draw_line (cairo_t	*cr,
					 PluginVals	*vals)
{
	LineObj *line = line_new (vals);
	gdouble cx, cy, in_radius, out_radius;

	get_global_vals (vals, &cx, &cy, &in_radius, &out_radius);

	cairo_save (cr);
	{
		double p1, psum;
		double max_len = 0;
		
		//compute radian for line_pitch
		switch (vals->line_pitch_type) {
			case LINE_PITCH_TYPE_IN:  max_len = 2 * G_PI * in_radius; break;
			case LINE_PITCH_TYPE_OUT: max_len = 2 * G_PI * out_radius; break;
		}
		p1 = G_PI * 2 * line->pitch / max_len;
		psum = 0;

		g_assert (p1 != 0);

		//normalize area center
		cairo_translate (cr, cx, cy);

		while (psum <= G_PI * 2) {
			double r = p1 - (p1 * g_rand_double (line->randobj_pitch) * line->pitch_rand / 100.0);

			cairo_save (cr);
			{
				//normalize line center
				cairo_translate (cr, 0, in_radius);
				cairo_rotate (cr, G_PI);

				line_draw (line, cr, in_radius - out_radius);
			}
			cairo_restore (cr);

			cairo_rotate (cr, r);
			psum += r;
		}
	}
	cairo_restore (cr);

	line_destroy (line);
}

gboolean
radiation_event_hook (cairo_t		*cr,
					  GdkEvent		*event,
					  PluginVals	*vals)
{
	static RadiationGuideObj dragging_obj = RADIATION_GUIDE_OBJ_NONE;
	static gdouble dragging_start_point_x = 0;
	static gdouble dragging_start_point_y = 0;

	switch (event->type) {
		case GDK_EXPOSE:
			break;

		case GDK_BUTTON_PRESS:
			dragging_start_point_x = event->button.x;
			dragging_start_point_y = event->button.y;
			dragging_obj = radiation_guide_hit_test (cr, event->any.window, vals, event->button.x, event->button.y);
			break;

		case GDK_BUTTON_RELEASE:
			dragging_start_point_x = 0;
			dragging_start_point_y = 0;
			dragging_obj = RADIATION_GUIDE_OBJ_NONE;
			break;

		case GDK_MOTION_NOTIFY:
			if (dragging_obj != RADIATION_GUIDE_OBJ_NONE) {
				gdouble gx, gy;
				gdouble cx, cy, r;
				gdouble resx, resy;
				gdouble temp;
				gimp_image_get_resolution (vals->image_ID, &resx, &resy);

				gx = event->motion.x;
				gy = event->motion.y;
				cairo_device_to_user (cr, &gx, &gy);

				switch (dragging_obj) {
					case RADIATION_GUIDE_OBJ_CENTER:
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_x), gimp_pixels_to_units (gx, vals->area_unit, resx));
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_y), gimp_pixels_to_units (gy, vals->area_unit, resy));
						break;
					case RADIATION_GUIDE_OBJ_IN_CIRCUM:
						cx = gimp_units_to_pixels (vals->radiation_x, vals->area_unit, resx);
						cy = gimp_units_to_pixels (vals->radiation_y, vals->area_unit, resy);
						r = sqrt (pow (gx - cx, 2) + pow (gy - cy, 2));
						temp = gimp_pixels_to_units (r, vals->area_unit, resx);
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_in_radius), temp);
						if (gtk_spin_button_get_value (GTK_SPIN_BUTTON (widgets.radiation_out_radius)) > temp) {
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_out_radius), temp);
						}
						break;
					case RADIATION_GUIDE_OBJ_OUT_CIRCUM:
						cx = gimp_units_to_pixels (vals->radiation_x, vals->area_unit, resx);
						cy = gimp_units_to_pixels (vals->radiation_y, vals->area_unit, resy);
						r = sqrt (pow (gx - cx, 2) + pow (gy - cy, 2));
						temp = gimp_pixels_to_units (r, vals->area_unit, resx);
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_out_radius), temp);
						if (gtk_spin_button_get_value (GTK_SPIN_BUTTON (widgets.radiation_in_radius)) < temp) {
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_in_radius), temp);
						}
						break;
				}
				gtk_widget_queue_draw (widgets.preview);
			}
			break;

		case GDK_KEY_PRESS:
			break;

		case GDK_KEY_RELEASE:
			break;

		default:
			break;
	}

	return FALSE;
}

/* vim: set fdm=marker ts=4 sw=4:*/
