/*
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2002,2004 Yeti (David Necas) <yeti@physics.muni.cz>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
   Version 1.0

   Nov 2002
   Yeti <yeti@physics.muni.cz>
   This is the first version.  No news, no changes, no known bugs.
*/

/*
   Version 1.1

   Nov 2003
   Yeti <yeti@physics.muni.cz>
   Bug fixes:
   * 100% is full image width/height when units are percents
   * works with gimp-1.3.21 and newer
*/

/* Version 1.2

   Jan 2004
   Yeti <yeti@physics.muni.cz>
   - changed dialog construction and main event loop to work with gimp-1.3.23
   - some code cleanup
*/

/* Version 1.3

   Apr 2004
   Yeti <yeti@physics.muni.cz>
   - A nonsuqare image mangling bug was fixed
   - Gimp-2.0 clean-up
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
   /* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "libgimp/stdplugins-intl.h"

#define PROCEDURE_NAME   "tileably_shrink"
#define DATA_KEY_VALS    "plug_in_tileably_shrink"

#define ENTRY_WIDTH 75

/* Don't change, though it's ugly */
#define SCALE_WIDTH 201

/* Plug-in settings. */
typedef struct
{
  gint     shrink_x;
  gint     shrink_y;
} TShrinkValues;

/* local prototypes */
static void query                  (void);
static void run                    (const gchar *name,
                                    gint nparams,
                                    const GimpParam *param,
                                    gint *nreturn_vals,
                                    GimpParam **return_vals);
static gint tshrink_dialog         (GimpDrawable *drawable,
                                    gint32 image_id);
static void tshrink                (GimpDrawable *drawable,
                                    gint32 image_id);
static void process_one_line       (GimpPixelRgn *src_rgn,
                                    GimpPixelRgn *dst_rgn,
                                    guchar *src,
                                    guchar *src2,
                                    guchar *dst,
                                    gint y);
static void process_one_line_alpha (GimpPixelRgn *src_rgn,
                                    GimpPixelRgn *dst_rgn,
                                    guchar *src,
                                    guchar *src2,
                                    guchar *dst,
                                    gint y);

/***** Local vars *****/

typedef struct
{
  GtkWidget *sizeentry;
} TShrinkControls;

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

static TShrinkValues tvals_defaults =
{
  10,   /* shrink_x */
  10    /* shrink_y */
};

static TShrinkValues tvals; /* = tvals_defaults; */
static TShrinkValues tvals_old;

gint wn, hn; /* new image dimensions */
gboolean has_alpha;
gdouble *xbeta, *ybeta; /* precomputed merging weights */

static TShrinkControls tctrl;

/***** Functions *****/

MAIN()

static void
query(void)
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
    { GIMP_PDB_INT32, "shrink_x", "Horizontal reduction" },
    { GIMP_PDB_INT32, "shrink_y", "Vertical reduction" },
  };

  gimp_install_procedure(DATA_KEY_VALS,
                         "Shinks an Image to make it it tileable",
                         "This plugin shrinks a given image to "
                         "make it tileable.",
                         "Yeti",
                         "Yeti",
                         "Nov 2002",
                         N_("<Image>/Filters/Map/Tileably Shrink..."),
                         "RGB*, GRAY*",
                         GIMP_PLUGIN,
                         G_N_ELEMENTS(args), 0,
                         args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam  *param,
    gint *nreturn_vals,
    GimpParam  **return_vals)
{
  static GimpParam values[1];
  GimpDrawable *drawable;
  GimpRunMode run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;

  run_mode = param[0].data.d_int32;

  /*  Initialize i18n support  */
  bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
  bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
  textdomain(GETTEXT_PACKAGE);

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;
  values[1].type = GIMP_PDB_IMAGE;

  /*  Get the specified drawable  */
  drawable = gimp_drawable_get(param[2].data.d_drawable);

  tvals = tvals_defaults;
  switch (run_mode) {
    case GIMP_RUN_INTERACTIVE:
      /*  Possibly retrieve data  */
      gimp_get_data(DATA_KEY_VALS, &tvals);

      /*  First acquire information with a dialog  */
      tvals_old = tvals;

      if (!tshrink_dialog(drawable, param[1].data.d_image)) {
        gimp_drawable_detach(drawable);
        return;
      }
      break;

    case GIMP_RUN_NONINTERACTIVE:
      /*  Make sure all the arguments are there!  */
      if (nparams != 5)
        status = GIMP_PDB_CALLING_ERROR;
      else {
        tvals.shrink_x = (gint)param[3].data.d_int32;
        tvals.shrink_y = (gint)param[4].data.d_int32;
      }
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      gimp_get_data(DATA_KEY_VALS, &tvals);
      break;

    default:
      break;
  }

  if (status == GIMP_PDB_SUCCESS) {
    /*  Make sure that the drawable is gray or RGB color  */
    if (gimp_drawable_is_rgb(drawable->drawable_id)
        || gimp_drawable_is_gray(drawable->drawable_id)) {
      gimp_progress_init(_("Tileably Shrink..."));

      tshrink(drawable, param[1].data.d_image);

      if (run_mode != GIMP_RUN_NONINTERACTIVE)
        gimp_displays_flush();

      /*  Store data  */
      if (run_mode == GIMP_RUN_INTERACTIVE)
        gimp_set_data(DATA_KEY_VALS, &tvals, sizeof(TShrinkValues));
    }
    else
      status = GIMP_PDB_EXECUTION_ERROR;
  }

  values[0].data.d_status = status;
  gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline void
yeti_table_attach_to_row(GtkWidget *table,
                         GtkWidget *child,
                         const gint row)
{
  gtk_table_attach(GTK_TABLE(table), child, 0, GTK_TABLE(table)->ncols,
                   row, row+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline GtkWidget*
yeti_frame_new_in_box(const gchar *name,
                      GtkWidget *box,
                      const gboolean expand,
                      const gboolean fill)
{
  GtkWidget *frame;

  frame = gtk_frame_new(name);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start(GTK_BOX(box), frame, expand, fill, 0);

  return frame;
}

static inline GtkWidget*
yeti_table_new_in_frame(const gint cols,
                        const gint rows,
                        GtkWidget *frame)
{
  GtkWidget *table;
  GtkWidget *align;

  align = gtk_alignment_new(0.5, 0.0, 1.0, 0.0);
  gtk_container_set_border_width(GTK_CONTAINER(align), 0);
  gtk_container_add(GTK_CONTAINER(frame), align);

  table = gtk_table_new(cols, rows, FALSE);
  gtk_table_set_col_spacings(GTK_TABLE(table), 4);
  gtk_table_set_row_spacings(GTK_TABLE(table), 2);
  gtk_container_set_border_width(GTK_CONTAINER(table), 4);
  gtk_container_add(GTK_CONTAINER(align), table);

  return table;
}

static inline void
yeti_progress_update(gint p,
                     gint max)
{
  double r;

  r = (double)p/max;
  gimp_progress_update(r);
}
/* }}} */

static gint
tshrink_dialog(GimpDrawable *drawable,
               gint32 image_id)
{
  GtkWidget *dlg;
  GtkWidget *frame;
  GtkWidget *table;
  gint       run;
  gint       width;
  gint       height;
  gdouble    xres;
  gdouble    yres;
  GimpUnit   unit;

  gimp_ui_init(PROCEDURE_NAME, TRUE);

  width = gimp_drawable_width(drawable->drawable_id);
  height = gimp_drawable_height(drawable->drawable_id);
  unit = gimp_image_get_unit(image_id);
  gimp_image_get_resolution(image_id, &xres, &yres);

  dlg = gimp_dialog_new(_("Tileably Shrink"), PROCEDURE_NAME,
                        NULL, 0,
                        gimp_standard_help_func, "filters/tileably-shrink.html",

                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                        GTK_STOCK_OK, GTK_RESPONSE_OK,
                        NULL);

  frame = yeti_frame_new_in_box(_("Parameter Settings"),
                                GTK_DIALOG(dlg)->vbox, TRUE, TRUE);
  table = yeti_table_new_in_frame(2, 1, frame);

  tctrl.sizeentry
    = gimp_coordinates_new(unit, "%a", TRUE, TRUE, 8,
                           GIMP_SIZE_ENTRY_UPDATE_SIZE,

                           1, TRUE, /* FIXME: z.constrain */

                           _("_Horizontal reduction:"),
                           1.0*tvals.shrink_x, 1.0*xres,
                           0.0, 0.5*width,
                           0.0, 1.0*width,

                           _("_Vertical reduction:"),
                           1.0*tvals.shrink_y, 1.0*yres,
                           0.0, 0.5*height,
                           0.0, 1.0*height);
  gtk_container_set_border_width(GTK_CONTAINER(tctrl.sizeentry), 4);
  gtk_table_set_row_spacing(GTK_TABLE(tctrl.sizeentry), 1, 4);
  yeti_table_attach_to_row(table, tctrl.sizeentry, 0);

  gtk_widget_show_all(dlg);

  run = gimp_dialog_run(GIMP_DIALOG(dlg));
  tvals.shrink_x
    = RINT(gimp_size_entry_get_refval(GIMP_SIZE_ENTRY(tctrl.sizeentry), 0));
  tvals.shrink_y
    = RINT(gimp_size_entry_get_refval(GIMP_SIZE_ENTRY(tctrl.sizeentry), 1));

  gtk_widget_destroy(dlg);

  return run == GTK_RESPONSE_OK;
}

static void
tshrink(GimpDrawable *drawable,
        gint32 image_id)
{
  gint32 shrinked_img, shrinked_lay;
  GimpDrawable *shrinked;
  GimpPixelRgn shrinked_rgn, src_rgn;
  guchar *dst, *src, *src2; /* row buffers */
  gint bpp, channels;
  GimpImageType imgtype;
  GimpImageBaseType imgbasetype;
  gint w, h, y, sx, sy;

  w = gimp_drawable_width(drawable->drawable_id);
  h = gimp_drawable_height(drawable->drawable_id);
  has_alpha = gimp_drawable_has_alpha(drawable->drawable_id);
  sx = tvals.shrink_x;
  wn = w - sx;
  sy = tvals.shrink_y;
  hn = h - sy;

  imgtype = gimp_drawable_type(drawable->drawable_id);
  imgbasetype = gimp_image_base_type(image_id);

  /* create a new image */
  shrinked_img = gimp_image_new(wn, hn, imgbasetype);
  shrinked_lay = gimp_layer_new(shrinked_img, "TShrinked", wn, hn,
                                imgtype, 100, GIMP_NORMAL_MODE);
  gimp_image_add_layer(shrinked_img, shrinked_lay, 0);
  shrinked = gimp_drawable_get(shrinked_lay);

  /* set up the pixel region stuff */
  gimp_tile_cache_ntiles((2*w + wn)/gimp_tile_width() + 2);
  gimp_pixel_rgn_init(&shrinked_rgn, shrinked, 0, 0, wn, hn, TRUE, TRUE);
  gimp_pixel_rgn_init(&src_rgn, drawable, 0, 0, w, hn, FALSE, FALSE);

  /* allocate row buffers (the resulting image is always of the same type) */
  channels = bpp = gimp_drawable_bpp(drawable->drawable_id);
  if (has_alpha)
    channels--;
  src = (guchar*)g_new(guchar, bpp*w);
  src2 = (guchar*)g_new(guchar, bpp*w);
  dst = (guchar*)g_new(guchar, bpp*wn);

  /* compute merging weights */
  if (sx > 0) {
    xbeta = g_new(gdouble, sx);
    for (y = 0; y < sx; y++)
      xbeta[y] = 0.5 - 0.5*cos(M_PI*(y + 1.0)/(sx + 1.0));
  }
  if (sy > 0) {
    ybeta = g_new(gdouble, sy);
    for (y = 0; y < sy; y++)
      ybeta[y] = 0.5 - 0.5*cos(M_PI*(y + 1.0)/(sy + 1.0));
  }

  /* render it! */
  for (y = 0; y < hn; y++) {
    /* read source rows */
    gimp_pixel_rgn_get_row(&src_rgn, src, 0, y + sy/2, w);
    if (y < sy - sy/2)
      gimp_pixel_rgn_get_row(&src_rgn, src2, 0, y + sy/2 + hn, w);
    if (y >= hn - sy/2)
      gimp_pixel_rgn_get_row(&src_rgn, src2, 0, y + sy/2 - hn, w);

    /* merge one line */
    if (has_alpha)
      process_one_line_alpha(&src_rgn, &shrinked_rgn, src, src2, dst, y);
    else
      process_one_line(&src_rgn, &shrinked_rgn, src, src2, dst, y);

    /* move result to the image */
    gimp_pixel_rgn_set_row(&shrinked_rgn, dst, 0, y, wn);
    yeti_progress_update(y, hn);
  }

  /* clean up */
  g_free(ybeta);
  g_free(xbeta);
  g_free(dst);
  g_free(src2);
  g_free(src);

  /* clean-up */
  gimp_drawable_flush(shrinked);
  gimp_drawable_merge_shadow(shrinked->drawable_id, TRUE);
  gimp_drawable_update(shrinked->drawable_id, 0, 0, wn, hn);
  gimp_drawable_detach(shrinked);
  gimp_display_new(shrinked_img);
}

/* This is ugly. The code could less than be half the size if we shrinked
 * the image toward 0,0 instead of its center.  But though the results are
 * exactly equivalent from the tilability point of view, the one-side's look
 * pretty queer.
 *
 * So we process nine different regions on the new image, which is moreover
 * shifted by tvals.shrink_?/2 against the original:
 *
 * +----+----------------------+-----+
 * | 8  |  6  .  7        . 8  |  6  |
 * |    |     .           .    |     |
 * +----o======================o-----+
 * |    H     .           .    H     |
 * | 3  H  1  .  2        . 3  H  1  |
 * |    H     .           .    H     |
 * |....H......................H.....|
 * |    H     .           .    H     |
 * |    H     .           .    H     |
 * | 5  H  4  .  0        . 5  H  4  |
 * |    H     .           .    H     |
 * |    H     .           .    H     |
 * |    H     .           .    H     |
 * |    H     .           .    H     |
 * |....H......................H.....|
 * | 8  H  6  .  7        . 8  H  6  |
 * |    H     .           .    H     |
 * +----o======================o-----+
 * |    |     .           .    |     |
 * | 3  |  1  .  2        . 3  |  1  |
 * |    |     .           .    |     |
 * +----+----------------------+-----+
 *
 * The outer is the original, the inner is the new image. Areas with the
 * same number are merged together, with corresponding weights to make it
 * smooth.
 */
static void
process_one_line(GimpPixelRgn *src_rgn,
                 GimpPixelRgn *dst_rgn,
                 guchar *src,
                 guchar *src2,
                 guchar *dst,
                 gint y)
{
  gint x;
  gint bpp = src_rgn->bpp;
  gint sx = tvals.shrink_x;
  gint sy = tvals.shrink_y;
  gint sx2 = sx/2;
  gint sy2 = sy/2;
  gint ch;
  gint wnbpp = wn*bpp;

  if (y < sy - sy2) {
    double qy = ybeta[sy - sy2 - y];

    for (x = 0; x < sx - sx2; x++) {
      double qx = xbeta[sx - sx2 - x];
      gint x2bpp = (x + sx2)*bpp;

      for (ch = 0; ch < bpp; ch++)
        dst[x*bpp + ch] = (1.0-qx)*(1.0-qy)*src[x2bpp + ch]
                          + qx*(1.0-qy)*src[x2bpp + wnbpp + ch]
                          + (1.0-qx)*qy*src2[x2bpp + ch]
                          + qx*qy*src2[x2bpp + wnbpp + ch];
    }
    for (x = sx - sx2; x < wn - sx2; x++) {
      gint x2bpp = (x + sx2)*bpp;

      for (ch = 0; ch < bpp; ch++)
        dst[x*bpp + ch] = (1.0-qy)*src[x2bpp + ch]
                          + qy*src2[x2bpp + ch];
    }
    for (x = wn - sx2; x < wn; x++) {
      double qx = xbeta[wn - 1 + sx2 - x];
      gint x2bpp = (x + sx2)*bpp;

      for (ch = 0; ch < bpp; ch++)
        dst[x*bpp + ch] = qx*(1.0-qy)*src[x2bpp + ch]
                          + (1.0-qx)*(1.0-qy)*src[x2bpp - wnbpp + ch]
                          + qx*qy*src2[x2bpp + ch]
                          + (1.0-qx)*qy*src2[x2bpp - wnbpp + ch];
    }
    return;
  }

  if (y > hn - sy2) {
    double qy = ybeta[hn - 1 + sy2 - y];

    for (x = 0; x < sx - sx2; x++) {
      double qx = xbeta[sx - sx2 - x];
      gint x2bpp = (x + sx2)*bpp;

      for (ch = 0; ch < bpp; ch++)
        dst[x*bpp + ch] = (1.0-qx)*qy*src[x2bpp + ch]
                          + qx*qy*src[x2bpp + wnbpp + ch]
                          + (1.0-qx)*(1.0-qy)*src2[x2bpp + ch]
                          + qx*(1.0-qy)*src2[x2bpp + wnbpp + ch];
    }
    for (x = sx - sx2; x < wn - sx2; x++) {
      gint x2bpp = (x + sx2)*bpp;

      for (ch = 0; ch < bpp; ch++)
        dst[x*bpp + ch] = qy*src[x2bpp + ch]
                          + (1.0-qy)*src2[x2bpp + ch];
    }
    for (x = wn - sx2; x < wn; x++) {
      double qx = xbeta[wn - 1 + sx2 - x];
      gint x2bpp = (x + sx2)*bpp;

      for (ch = 0; ch < bpp; ch++)
        dst[x*bpp + ch] = qx*qy*src[x2bpp + ch]
                          + (1.0-qx)*qy*src[x2bpp - wnbpp + ch]
                          + qx*(1.0-qy)*src2[x2bpp + ch]
                          + (1.0-qx)*(1.0-qy)*src2[x2bpp - wnbpp + ch];
    }
    return;
  }

  for (x = 0; x < sx - sx2; x++) {
    double qx = xbeta[sx - sx2 - x];
    gint x2bpp = (x + sx2)*bpp;

    for (ch = 0; ch < bpp; ch++)
      dst[x*bpp + ch] = (1.0-qx)*src[x2bpp + ch]
                        + qx*src[x2bpp + wnbpp + ch];
  }
  memcpy(dst + sx2*bpp, src + sx*bpp, bpp*(wn - sx));
  for (x = wn - sx2; x < wn; x++) {
    double qx = xbeta[wn - 1 + sx2 - x];
    gint x2bpp = (x + sx2)*bpp;

    for (ch = 0; ch < bpp; ch++)
      dst[x*bpp + ch] = qx*src[x2bpp + ch]
                        + (1.0-qx)*src[x2bpp - wnbpp + ch];
  }
}

/* Mergin' with alpha channels and with different weights is just a mess.
 * The equations are:
 *
 *   alpha  =  alpha_1 * weight_1  +  alpha_2 * weight_2
 *
 * for alpha, and
 *
 *   ch  =  (alpha_1 * weight_1 * ch_1  +  alpha_2 * weight_2 * ch_2) / alpha
 *
 * for color channels when alpha <> 0.  When alpha == 0 the color is
 * arbitrary (I choosed black, it's the easiest to implement). */
static void
process_one_line_alpha(GimpPixelRgn *src_rgn,
                       GimpPixelRgn *dst_rgn,
                       guchar *src,
                       guchar *src2,
                       guchar *dst,
                       gint y)
{
  gint x;
  gint bpp = src_rgn->bpp;
  gint sx = tvals.shrink_x;
  gint sy = tvals.shrink_y;
  gint sx2 = sx/2;
  gint sy2 = sy/2;
  gint ch;
  gint wnbpp = wn*bpp;

  if (y < sy - sy2) {
    double qy = ybeta[sy - sy2 - y];

    for (x = 0; x < sx - sx2; x++) {
      double qx = xbeta[sx - sx2 - x];
      gint x2bpp = (x + sx2)*bpp;
      gint alphaA = src[x2bpp + bpp-1];
      gint alphaB1 = src[x2bpp + wnbpp + bpp-1];
      gint alphaB2 = src2[x2bpp + bpp-1];
      gint alphaC = src2[x2bpp + wnbpp + bpp-1];
      gdouble alpha;

      alpha = (1.0-qx)*(1.0-qy)*alphaA
              + qx*(1.0-qy)*alphaB1
              + (1.0-qx)*qy*alphaB2
              + qx*qy*alphaC;
      dst[x*bpp + bpp-1] = alpha;
      if (alpha == 0.0)
        alpha = 1.0;

      for (ch = 0; ch < bpp-1; ch++)
        dst[x*bpp + ch] = ((1.0-qx)*(1.0-qy)*src[x2bpp + ch]*alphaA
                           + qx*(1.0-qy)*src[x2bpp + wnbpp + ch]*alphaB1
                           + (1.0-qx)*qy*src2[x2bpp + ch]*alphaB2
                           + qx*qy*src2[x2bpp + wnbpp + ch]*alphaC)
                          /alpha;
    }
    for (x = sx - sx2; x < wn - sx2; x++) {
      gint x2bpp = (x + sx2)*bpp;
      gint alphaA = src[x2bpp + bpp-1];
      gint alphaB2 = src2[x2bpp + bpp-1];
      gdouble alpha;

      alpha = (1.0-qy)*alphaA
              + qy*alphaB2;
      dst[x*bpp + bpp-1] = alpha;
      if (alpha == 0.0)
        alpha = 1.0;

      for (ch = 0; ch < bpp-1; ch++)
        dst[x*bpp + ch] = ((1.0-qy)*src[x2bpp + ch]*alphaA
                           + qy*src2[x2bpp + ch]*alphaB2)
                          /alpha;
    }
    for (x = wn - sx2; x < wn; x++) {
      double qx = xbeta[wn - 1 + sx2 - x];
      gint x2bpp = (x + sx2)*bpp;
      gint alphaA = src[x2bpp + bpp-1];
      gint alphaB1 = src[x2bpp - wnbpp + bpp-1];
      gint alphaB2 = src2[x2bpp + bpp-1];
      gint alphaC = src2[x2bpp - wnbpp + bpp-1];
      gdouble alpha;

      alpha = qx*(1.0-qy)*alphaA
              + (1.0-qx)*(1.0-qy)*alphaB1
              + qx*qy*alphaB2
              + (1.0-qx)*qy*alphaC;
      dst[x*bpp + bpp-1] = alpha;
      if (alpha == 0.0)
        alpha = 1.0;

      for (ch = 0; ch < bpp-1; ch++)
        dst[x*bpp + ch] = (qx*(1.0-qy)*src[x2bpp + ch]*alphaA
                           + (1.0-qx)*(1.0-qy)*src[x2bpp - wnbpp + ch]*alphaB1
                           + qx*qy*src2[x2bpp + ch]*alphaB2
                           + (1.0-qx)*qy*src2[x2bpp - wnbpp + ch]*alphaC)
                          /alpha;
    }
    return;
  }

  if (y > hn - sy2) {
    double qy = ybeta[hn - 1 + sy2 - y];

    for (x = 0; x < sx - sx2; x++) {
      double qx = xbeta[sx - sx2 - x];
      gint x2bpp = (x + sx2)*bpp;
      gint alphaA = src[x2bpp + bpp-1];
      gint alphaB1 = src[x2bpp + wnbpp + bpp-1];
      gint alphaB2 = src2[x2bpp + bpp-1];
      gint alphaC = src2[x2bpp + wnbpp + bpp-1];
      gdouble alpha;

      alpha = (1.0-qx)*qy*alphaA
              + qx*qy*alphaB1
              + (1.0-qx)*(1.0-qy)*alphaB2
              + qx*(1.0-qy)*alphaC;
      dst[x*bpp + bpp-1] = alpha;
      if (alpha == 0.0)
        alpha = 1.0;

      for (ch = 0; ch < bpp-1; ch++)
        dst[x*bpp + ch] = ((1.0-qx)*qy*src[x2bpp + ch]*alphaA
                           + qx*qy*src[x2bpp + wnbpp + ch]*alphaB1
                           + (1.0-qx)*(1.0-qy)*src2[x2bpp + ch]*alphaB2
                           + qx*(1.0-qy)*src2[x2bpp + wnbpp + ch]*alphaC)
                          /alpha;
    }
    for (x = sx - sx2; x < wn - sx2; x++) {
      gint x2bpp = (x + sx2)*bpp;
      gint alphaA = src[x2bpp + bpp-1];
      gint alphaB2 = src2[x2bpp + bpp-1];
      gdouble alpha;

      alpha = qy*alphaA
              + (1.0-qy)*alphaB2;
      dst[x*bpp + bpp-1] = alpha;
      if (alpha == 0.0)
        alpha = 1.0;

      for (ch = 0; ch < bpp-1; ch++)
        dst[x*bpp + ch] = (qy*src[x2bpp + ch]*alphaA
                           + (1.0-qy)*src2[x2bpp + ch]*alphaB2)
                          /alpha;
    }
    for (x = wn - sx2; x < wn; x++) {
      double qx = xbeta[wn - 1 + sx2 - x];
      gint x2bpp = (x + sx2)*bpp;
      gint alphaA = src[x2bpp + bpp-1];
      gint alphaB1 = src[x2bpp - wnbpp + bpp-1];
      gint alphaB2 = src2[x2bpp + bpp-1];
      gint alphaC = src2[x2bpp - wnbpp + bpp-1];
      gdouble alpha;

      alpha = qx*qy*alphaA
              + (1.0-qx)*qy*alphaB1
              + qx*(1.0-qy)*alphaB2
              + (1.0-qx)*(1.0-qy)*alphaC;
      dst[x*bpp + bpp-1] = alpha;
      if (alpha == 0.0)
        alpha = 1.0;

      for (ch = 0; ch < bpp-1; ch++)
        dst[x*bpp + ch] = (qx*qy*src[x2bpp + ch]*alphaA
                           + (1.0-qx)*qy*src[x2bpp - wnbpp + ch]*alphaB1
                           + qx*(1.0-qy)*src2[x2bpp + ch]*alphaB2
                           + (1.0-qx)*(1.0-qy)*src2[x2bpp - wnbpp + ch]*alphaC)
                          /alpha;
    }
    return;
  }

  for (x = 0; x < sx - sx2; x++) {
    double qx = xbeta[sx - sx2 - x];
    gint x2bpp = (x + sx2)*bpp;
    gint alphaA = src[x2bpp + bpp-1];
    gint alphaB1 = src[x2bpp + wnbpp + bpp-1];
    gdouble alpha;

    alpha = (1.0-qx)*alphaA
            + qx*alphaB1;
    dst[x*bpp + bpp-1] = alpha;
    if (alpha == 0.0)
      alpha = 1.0;

    for (ch = 0; ch < bpp-1; ch++)
      dst[x*bpp + ch] = ((1.0-qx)*src[x2bpp + ch]*alphaA
                         + qx*src[x2bpp + wnbpp + ch]*alphaB1)
                        /alpha;
  }
  memcpy(dst + sx2*bpp, src + sx*bpp, bpp*(wn - sx));
  for (x = wn - sx2; x < wn; x++) {
    double qx = xbeta[wn - 1 + sx2 - x];
    gint x2bpp = (x + sx2)*bpp;
    gint alphaA = src[x2bpp + bpp-1];
    gint alphaB1 = src[x2bpp - wnbpp + bpp-1];
    gdouble alpha;

    alpha = qx*alphaA
            + (1.0-qx)*alphaB1;
    dst[x*bpp + bpp-1] = alpha;
    if (alpha == 0.0)
      alpha = 1.0;

    for (ch = 0; ch < bpp-1; ch++)
      dst[x*bpp + ch] = (qx*src[x2bpp + ch]*alphaA
                         + (1.0-qx)*src[x2bpp - wnbpp + ch]*alphaB1)
                         /alpha;
  }
}
